#ifndef FACTORIZATIONCONTROLLER_H
#define FACTORIZATIONCONTROLLER_H

#include <QObject>
#include <QTimer>
#include <QVariant>
#include <qqmlintegration.h>
#include <qtmetamacros.h>

/**
 * @class FactorizationController
 * @brief A class that performs integer factorization in a step-by-step manner.
 *
 * This class uses a timer to progressively factorize a given number, allowing
 * for real-time updates on progress and discovered factors. The computation will
 * run in a single thread, utilizing 0 tick timers to perform a small number of
 * iterations on each cycle of the event loop.
 *
 * It is designed to be used in a Qt-based QML application.
 */
class FactorizationController : public QObject {
    Q_OBJECT
    QML_ELEMENT
    Q_PROPERTY(bool isRunning READ isRunning NOTIFY isRunningChanged)
    Q_PROPERTY(bool isFinished READ isFinished NOTIFY isFinishedChanged)
    Q_PROPERTY(bool isPaused READ isPaused NOTIFY isPausedChanged)
    Q_PROPERTY(int iterationsPerCycle READ iterationsPerCycle WRITE setIterationsPerCycle NOTIFY iterationsPerCycleChanged)
    Q_PROPERTY(bool useSqrtOptimization READ useSqrtOptimization WRITE setUseSqrtOptimization NOTIFY useSqrtOptimizationChanged)
    Q_PROPERTY(bool pauseOnFound READ pauseOnFound WRITE setPauseOnFound NOTIFY pauseOnFoundChanged)
    Q_PROPERTY(unsigned long long number READ number NOTIFY numberChanged)
    Q_PROPERTY(unsigned long long curFactNumber READ curFactNumber NOTIFY curFactNumberChanged)
    Q_PROPERTY(unsigned long long currentFactor READ currentFactor NOTIFY currentFactorChanged)
    Q_PROPERTY(unsigned long long stopFactor READ stopFactor NOTIFY stopFactorChanged)
    Q_PROPERTY(int progress READ progress NOTIFY progressChanged)
    Q_PROPERTY(QList<unsigned long long> factors READ factors NOTIFY factorsChanged)

  public:
    explicit FactorizationController(QObject* parent = nullptr);

    // region: Getters

    /**
     * @brief Gets the current progress of factorization.
     * @return The progress percentage (0-100).
     */
    int progress() const;

    /**
     * @brief Checks if the factorization process is running.
     * @return True if the factorization is in progress, otherwise false.
     *
     * This just informs whether the task is active (scheduled to run), not whether it's
     * running at this very moment, as this implementation doesn't use threads.
     */
    bool isRunning() const;

    /**
     * @brief Checks if the factorization process is done.
     * @return True if the factorization is completed, otherwise false.
     *
     * This will only be true if the factorization process finished without being
     * explicitly stopped ahead of time (e.g. when the computation is done and we
     * have all the factors).
     */
    bool isFinished() const;

    /**
     * @brief Checks if the factorization process is currently paused.
     * @return True if the factorization process is paused, otherwise false.
     *
     * This function differs from just !isRunning, as this will only be set when
     * the computation is currently paused, but can still be resumed. E.g. the
     * computation was already started at some point, and stopped (paused) later,
     * without having been reset.
     *
     * This being true is a prerequisite to resuming.
     */
    bool isPaused() const;

    /**
     * @brief Gets the number of iterations to be performed per cycle.
     * @return The number of iterations performed per cycle.
     */
    int iterationsPerCycle() const;

    /**
     * @brief Checks whether sqrt optimization is enabled.
     * @return True if the sqrt optimization is enabled, otherwise False.
     */
    bool useSqrtOptimization() const;

    /**
     * @brief Checks whether the computation should be paused once a new factor is found.
     * @return True if the computation will get paused on new factor being found, otherwise False.
     */
    bool pauseOnFound() const;

    /**
     * @brief Gets the original number being factorized.
     * @return The number being factorized.
     */
    unsigned long long number() const;

    /**
     * @brief The current number being factorized.
     * @return The current number being factorized.
     *
     * As the factorization process is going on, the number being factorized might change
     * from the original number, as we're dividing it by the found factors.
     */
    unsigned long long curFactNumber() const;

    /**
     * @brief Gets the current factor being tested.
     * @return The current factor.
     *
     * The factorization process is implemented in a way that simply checks whether a number
     * is a factor, going one by one. This returns the current factor (to be checked in the
     * next iteration).
     */
    unsigned long long currentFactor() const;

    /**
     * @brief Get the number at which the factorization process will stop
     * @return The current stop factor.
     *
     * The stop factor determines at which factor will the factorization process be stopped.
     * If sqrt optimizations are enabled, this will generally be the square root of the number
     * currently being factorized. If not, it will be the number being factorized - 1.
     */
    unsigned long long stopFactor() const;

    /**
     * @brief Gets the list of discovered factors (may not be complete).
     * @return A QList of extracted prime factors.
     *
     * This returns the factors that were already discovered Note that this list might not
     * be complete (if the computation is still running). If you need to know whether you
     * can trust this result, check isFinished first.
     */
    QList<unsigned long long> factors() const;

    // endregion
    // region: Setters

    /**
     * @brief Set the number of iterations per cycle.
     *
     * This is a way to speed up the computation, by increasing the amount of iterations
     * performed per each computation (on every event loop cycle). Setting this to a number
     * too high might lead to OS thinking the application is not responding, and the UI
     * feeling slow. The optimal number depends on the speed of the CPU and the allocated
     * CPU time to this process.
     *
     * This value can be changed even as a computation is ongoing.
     */
    void setIterationsPerCycle(int number);

    /**
     * @brief Enable or disable using sqrt optimizations.
     *
     * When sqrt optimization is enabled, only factors up to sqrt of the original number
     * will be checked. This allows the algorithm to finish quicker, while still finding
     * all of the factors (except the last one, which will be the original number after dividing
     * it by all of the found factors, guaranteed to be prime or 1).
     *
     * Generally, there's no reason to disable this optimization, however, if you wish to
     * perform a slower search, it is a way to achieve that.
     *
     * Note that changing this value while the computation is in progress will reset the
     * computation.
     */
    void setUseSqrtOptimization(bool value);

    /**
     * @brief Enable or disable auto-pausing on new factor being found.
     *
     * This value can be changed even as a computation is ongoing.
     */
    void setPauseOnFound(bool value);

    // endregion
    // region: Functions

    /**
     * @brief Starts the factorization process for a given number.
     * @param number The number to be factorized.
     *
     * This will start a 0-tick timer, which triggers a partial computation on each run,
     * to let the event loop cycle in between. This implementation does NOT rely on threads.
     */
    Q_INVOKABLE void start(unsigned long long number);

    /**
     * @brief Stops the ongoing factorization process.
     *
     * This will stop the timer, meaning the computation will no longer get triggered
     * by the event loop, essentially pausing it. Note that this does not reset the already
     * computed details, which means the computation can be resumed later, allowing for
     * inspection of the currently computed values.
     */
    Q_INVOKABLE void stop();

    /**
     * @brief Resumes a previously stopped factorization process.
     */
    Q_INVOKABLE void resume();

    /**
     * @brief Resets the factorization process data.
     *
     * This will clear the current progress (if any). If the factorization process is running,
     * this will stop it. If the factorization process is paused, this will reset isPaused,
     * forcing a restart.
     */
    Q_INVOKABLE void reset();

    // endregion

  signals:
    void isRunningChanged();
    void isFinishedChanged();
    void isPausedChanged();
    void useSqrtOptimizationChanged();
    void pauseOnFoundChanged();
    void iterationsPerCycleChanged();
    void numberChanged();
    void currentFactorChanged();
    void factorsChanged();
    void progressChanged();
    void curFactNumberChanged();
    void stopFactorChanged();

  private slots:
    void onTimerTick();

  private:
    bool m_isRunning; ///< Indicates whether factorization is active.
    bool m_isFinished; ///< Indicates whether the factorization process is done.
    bool m_isPaused; ///< Indicates whether the factorization process is paused (can be resumed).
    bool m_useSqrtOptimization; ///< Indicates whether to use the sqrt optimization
    bool m_pauseOnFound; ///< Indicates whether the computation should be paused when a new factor is found.
    int m_iterationsPerCycle; ///< The number of iterations to perform per cycle.
    unsigned long long m_currentFactNumber; ///< The number currently being factorized.
    unsigned long long m_originalNumber; ///< The original input number.
    unsigned long long m_currentFactor; ///< The current divisor being checked.
    unsigned long long m_stopFactor; ///< The stopping limit (square root of original number, saved for efficiency).
    QList<unsigned long long> m_factors; ///< List of discovered prime factors.
    QTimer m_timer; ///< Timer for stepwise execution.

    /**
     * @brief Performs one step of the factorization process.
     */
    void factorize();
};

#endif // FACTORIZATIONCONTROLLER_H
